@@ -4,14 +4,20 @@ For the sake of sanity, we only do internal formatting with unicode, which can
 be returned by the __unicode__ special method. We use ReprMixin to provide the
 __repr__ method so that things can work on Python 2.
 """
+from __future__ import absolute_import
+from __future__ import division
+from __future__ import print_function
+import contextlib
 from datetime import datetime, timedelta
 import functools
 
 import numpy as np
 import pandas as pd
+from pandas.tslib import OutOfBoundsDatetime
 
 from .options import OPTIONS
-from .pycompat import PY2, iteritems, unicode_type, bytes_type, dask_array_type
+from .pycompat import (
+    PY2, unicode_type, bytes_type, dask_array_type, OrderedDict, basestring)
 
 
 def pretty_print(x, numchars):
@@ -83,9 +89,24 @@ def first_n_items(x, n_desired):
     return np.asarray(x).flat[:n_desired]
 
 
+def last_item(x):
+    """Returns the last item of an array in a list or an empty list."""
+    if x.size == 0:
+        # work around for https://github.com/numpy/numpy/issues/5195
+        return []
+
+    indexer = (slice(-1, None),) * x.ndim
+    return np.ravel(x[indexer]).tolist()
+
+
 def format_timestamp(t):
     """Cast given object to a Timestamp and return a nicely formatted string"""
-    datetime_str = unicode_type(pd.Timestamp(t))
+    # Timestamp is only valid for 1678 to 2262
+    try:
+        datetime_str = unicode_type(pd.Timestamp(t))
+    except OutOfBoundsDatetime:
+        datetime_str = unicode_type(t)
+
     try:
         date_str, time_str = datetime_str.split()
     except ValueError:
@@ -186,9 +207,24 @@ def _summarize_var_or_coord(name, var, col_width, show_values=True,
         values_str = format_array_flat(var, max_width - len(front_str))
     else:
         values_str = u'...'
+
     return front_str + values_str
 
 
+def _summarize_coord_multiindex(coord, col_width, marker):
+    first_col = pretty_print(u'  %s %s ' % (marker, coord.name), col_width)
+    return u'%s(%s) MultiIndex' % (first_col, unicode_type(coord.dims[0]))
+
+
+def _summarize_coord_levels(coord, col_width, marker=u'-'):
+    relevant_coord = coord[:30]
+    return u'\n'.join(
+        [_summarize_var_or_coord(lname,
+                                 relevant_coord.get_level_variable(lname),
+                                 col_width, marker=marker)
+         for lname in coord.level_names])
+
+
 def _not_remote(var):
     """Helper function to identify if array is positively identifiable as
     coming from a remote source.
@@ -208,6 +244,12 @@ def summarize_coord(name, var, col_width):
     is_index = name in var.dims
     show_values = is_index or _not_remote(var)
     marker = u'*' if is_index else u' '
+    if is_index:
+        coord = var.variable.to_index_variable()
+        if coord.level_names is not None:
+            return u'\n'.join(
+                [_summarize_coord_multiindex(coord, col_width, marker),
+                 _summarize_coord_levels(coord, col_width)])
     return _summarize_var_or_coord(name, var, col_width, show_values, marker)
 
 
@@ -219,9 +261,26 @@ def summarize_attr(key, value, col_width=None):
 EMPTY_REPR = u'    *empty*'
 
 
-def _calculate_col_width(mapping):
-    max_name_length = (max(len(unicode_type(k)) for k in mapping)
-                       if mapping else 0)
+def _get_col_items(mapping):
+    """Get all column items to format, including both keys of `mapping`
+    and MultiIndex levels if any.
+    """
+    from .variable import IndexVariable
+
+    col_items = []
+    for k, v in mapping.items():
+        col_items.append(k)
+        var = getattr(v, 'variable', v)
+        if isinstance(var, IndexVariable):
+            level_names = var.to_index_variable().level_names
+            if level_names is not None:
+                col_items += list(level_names)
+    return col_items
+
+
+def _calculate_col_width(col_items):
+    max_name_length = (max(len(unicode_type(s)) for s in col_items)
+                       if col_items else 0)
     col_width = max(max_name_length, 7) + 6
     return col_width
 
@@ -237,18 +296,21 @@ def _mapping_repr(mapping, title, summarizer, col_width=None):
     return u'\n'.join(summary)
 
 
-coords_repr = functools.partial(_mapping_repr, title=u'Coordinates',
-                                summarizer=summarize_coord)
-
-
-vars_repr = functools.partial(_mapping_repr, title=u'Data variables',
-                              summarizer=summarize_var)
+data_vars_repr = functools.partial(_mapping_repr, title=u'Data variables',
+                                   summarizer=summarize_var)
 
 
 attrs_repr = functools.partial(_mapping_repr, title=u'Attributes',
                                summarizer=summarize_attr)
 
 
+def coords_repr(coords, col_width=None):
+    if col_width is None:
+        col_width = _calculate_col_width(_get_col_items(coords))
+    return _mapping_repr(coords, title=u'Coordinates',
+                         summarizer=summarize_coord, col_width=col_width)
+
+
 def indexes_repr(indexes):
     summary = []
     for k, v in indexes.items():
@@ -256,22 +318,62 @@ def indexes_repr(indexes):
     return u'\n'.join(summary)
 
 
+def dim_summary(obj):
+    elements = [u'%s: %s' % (k, v) for k, v in obj.sizes.items()]
+    return u', '.join(elements)
+
+
+def unindexed_dims_repr(dims, coords):
+    unindexed_dims = [d for d in dims if d not in coords]
+    if unindexed_dims:
+        dims_str = u', '.join(u'%s' % d for d in unindexed_dims)
+        return u'Unindexed dimensions:\n' + u' ' * 4 + dims_str
+    else:
+        return None
+
+
+@contextlib.contextmanager
+def set_numpy_options(*args, **kwargs):
+    original = np.get_printoptions()
+    np.set_printoptions(*args, **kwargs)
+    yield
+    np.set_printoptions(**original)
+
+
+def short_array_repr(array):
+    array = np.asarray(array)
+    # default to lower precision so a full (abbreviated) line can fit on
+    # one line with the default display_width
+    options = {
+        'precision': 6,
+        'linewidth': OPTIONS['display_width'],
+        'threshold': 200,
+    }
+    if array.ndim < 3:
+        edgeitems = 3
+    elif array.ndim == 3:
+        edgeitems = 2
+    else:
+        edgeitems = 1
+    options['edgeitems'] = edgeitems
+    with set_numpy_options(**options):
+        return repr(array)
+
+
 def array_repr(arr):
-    # used for DataArray, Variable and Coordinate
+    # used for DataArray, Variable and IndexVariable
     if hasattr(arr, 'name') and arr.name is not None:
         name_str = '%r ' % arr.name
     else:
         name_str = u''
-    dim_summary = u', '.join(u'%s: %s' % (k, v) for k, v
-                            in zip(arr.dims, arr.shape))
 
     summary = [u'<xarray.%s %s(%s)>'
-               % (type(arr).__name__, name_str, dim_summary)]
+               % (type(arr).__name__, name_str, dim_summary(arr))]
 
     if isinstance(getattr(arr, 'variable', arr)._data, dask_array_type):
         summary.append(repr(arr.data))
     elif arr._in_memory or arr.size < 1e5:
-        summary.append(repr(arr.values))
+        summary.append(short_array_repr(arr.values))
     else:
         summary.append(u'[%s values with dtype=%s]' % (arr.size, arr.dtype))
 
@@ -279,6 +381,10 @@ def array_repr(arr):
         if arr.coords:
             summary.append(repr(arr.coords))
 
+        unindexed_dims_str = unindexed_dims_repr(arr.dims, arr.coords)
+        if unindexed_dims_str:
+            summary.append(unindexed_dims_str)
+
     if arr.attrs:
         summary.append(attrs_repr(arr.attrs))
 
@@ -288,14 +394,19 @@ def array_repr(arr):
 def dataset_repr(ds):
     summary = [u'<xarray.%s>' % type(ds).__name__]
 
-    col_width = _calculate_col_width(ds)
+    col_width = _calculate_col_width(_get_col_items(ds))
 
     dims_start = pretty_print(u'Dimensions:', col_width)
-    all_dim_strings = [u'%s: %s' % (k, v) for k, v in iteritems(ds.dims)]
-    summary.append(u'%s(%s)' % (dims_start, ', '.join(all_dim_strings)))
+    summary.append(u'%s(%s)' % (dims_start, dim_summary(ds)))
 
     summary.append(coords_repr(ds.coords, col_width=col_width))
-    summary.append(vars_repr(ds.data_vars, col_width=col_width))
+
+    unindexed_dims_str = unindexed_dims_repr(ds.dims, ds.coords)
+    if unindexed_dims_str:
+        summary.append(unindexed_dims_str)
+
+    summary.append(data_vars_repr(ds.data_vars, col_width=col_width))
+
     if ds.attrs:
         summary.append(attrs_repr(ds.attrs))
 
